/**
 * @file TermSharing.hpp
 * Defines class TermSharing.
 *
 * @since 28/12/2007 Manchester
 */

#ifndef __TermSharing__
#define __TermSharing__

#include "Lib/Set.hpp"
#include "Kernel/Term.hpp"

using namespace Lib;
using namespace Kernel;

namespace Indexing {

class TermSharing
{
public:
  TermSharing();
  ~TermSharing();

  Term* insert(Term*);
  Term* insertRecurrently(Term*);

  Literal* insert(Literal*);
  Literal* insertVariableEquality(Literal* lit,unsigned sort);

  Literal* tryGetOpposite(Literal* l);

  /** The hash function of this literal */
  inline static unsigned hash(const Literal* l)
  { return l->hash(); }
  /** The hash function of this term */
  inline static unsigned hash(const Term* t)
  { return t->hash(); }
  static bool equals(const Term* t1,const Term* t2);

  static bool equals(const Literal* l1, const Literal* l2, bool opposite=false);

  struct OpLitWrapper {
    OpLitWrapper(Literal* l) : l(l) {}
    Literal* l;
  };
  inline static unsigned hash(const OpLitWrapper& w)
  { return w.l->oppositeHash(); }
  static bool equals(const Literal* l1,const OpLitWrapper& w) {
    return equals(l1, w.l, true);
  }

private:
  bool argNormGt(TermList t1, TermList t2);

  /** The set storing all terms */
  Set<Term*,TermSharing> _terms;
  /** The set storing all literals */
  Set<Literal*,TermSharing> _literals;
  /** Number of terms stored */
  unsigned _totalTerms;
  /** Number of ground terms stored */
  unsigned _groundTerms;
  /** Number of literals stored */
  unsigned _totalLiterals;
  /** Number of ground literals stored */
  unsigned _groundLiterals;
  /** Number of literal insertions */
  unsigned _literalInsertions;
  /** Number of term insertions */
  unsigned _termInsertions;
}; // class TermSharing

} // namespace Indexing

#endif
