open Lib
(* index is based on feature index of S. Schulz  *)

type clause = Clause.clause
(* type literal = Clause.literal *)
(* type var = Var.var *)
(* type lit = Term.literal *)

(*type symbol = Symbol.symbol 

module Key = 
  struct
    type t      = symbol
    let equal   = (==)  
    let hash    = get_fast_key 
    let compare = compare_fast_key
  end
*)
module SymbMap = Symbol.Map

module Set = Clause.Set

module type Feature =
sig
        type t
        val compare : t -> t -> int
(* val get_feature_list : clause -> t list *)
end

(* Get Var Count: *)
(* 1) Generate mapping between var and the number of its containing clauses *)
(* 2) Generate mapping between var and the number of containing clauses with more than one instantiation *)

(* Main loop *)
(* Find all clauses that hold the variable in positive polarity *)
(* Find all clauses that hold the variable in negative polarity *)
(* Find resolution and apply on N P *)

type pos_neg_cl_lists =
	{ pos_cl_list   : clause list;
          neg_cl_list   : clause list;
	  multi_occ_cnt : int }


let rec add_clause_pred_map cl pred_map =
	let f curr_pred_map lit =
		let atom = Term.get_atom lit in
		let is_neg = Term.is_neg_lit lit in
		let pred_symb = Term.get_top_symb atom in
		let pred_list = Clause.find_all (fun l_pred -> ((Term.get_top_symb (Term.get_atom l_pred)) == pred_symb)) cl in
		let curr_occ = List.length pred_list in
		let curr_cnt = 
			if (curr_occ > 1) then
				1
			else
			    	0;
		in
		try
			let old_pos_neg = SymbMap.find pred_symb curr_pred_map in
			let new_pos_neg =
				if is_neg then
					{ old_pos_neg with
						neg_cl_list = cl:: (old_pos_neg.neg_cl_list);
						multi_occ_cnt = old_pos_neg.multi_occ_cnt + curr_cnt
					}
				else
					{ old_pos_neg with
						pos_cl_list = cl:: (old_pos_neg.pos_cl_list);
						multi_occ_cnt = old_pos_neg.multi_occ_cnt + curr_cnt
					}
			in
			SymbMap.add pred_symb new_pos_neg curr_pred_map
		with
			Not_found ->
				let new_pos_neg =
					if is_neg then
						{ pos_cl_list = [];
						  neg_cl_list = [cl];
						  multi_occ_cnt = curr_cnt;	  
						}
					else
						{ pos_cl_list = [cl];
						  neg_cl_list = [];
						  multi_occ_cnt = curr_cnt;
						}
				in
				SymbMap.add pred_symb new_pos_neg curr_pred_map
	in
	let lit_list = Clause.get_literals cl in
	List.fold_left f pred_map lit_list





let rec remove_clause_pred_map pred_map cl =
	let f curr_pred_map lit =
		let atom = Term.get_atom lit in
		let is_neg = Term.is_neg_lit lit in
		let pred_symb = Term.get_top_symb atom in
		let pred_list = Clause.find_all (fun l_pred -> ((Term.get_top_symb (Term.get_atom l_pred)) == pred_symb)) cl in
		let curr_occ = List.length pred_list in
		let curr_cnt = 
			if (curr_occ > 1) then
				1
			else
			    	0;
		in
		let same_clause c2 = Clause.equal cl c2 in
		try
			let old_pos_neg = SymbMap.find pred_symb curr_pred_map in
			let new_pos_neg =
				if is_neg then
					{ old_pos_neg with
						neg_cl_list = snd (List.partition same_clause old_pos_neg.neg_cl_list);
						multi_occ_cnt = old_pos_neg.multi_occ_cnt - curr_cnt
					}
				else
					{ old_pos_neg with
						pos_cl_list = snd (List.partition same_clause old_pos_neg.pos_cl_list);
						multi_occ_cnt = old_pos_neg.multi_occ_cnt - curr_cnt
					}
			in
			SymbMap.add pred_symb new_pos_neg curr_pred_map
		with
			Not_found ->
				SymbMap.empty
	in
	let lit_list = Clause.get_literals cl in
	List.fold_left f pred_map lit_list


let resolution c1 l1 compl_l1 c_list2 = 
  let new_litlist1 = 
    Clause.find_all (fun lit -> not(l1 == lit)) c1 
  in 
  let f rest c2 = 
    let new_litlist2 = 
      Clause.find_all (fun lit -> not(compl_l1 == lit)) c2 in 
    let conclusion = 
                Clause.create (new_litlist1@new_litlist2) in
    (* Q: what is the purpose of the function below? *)
    Clause.assign_tstp_source_resolution conclusion [c1;c2] [l1;compl_l1];
    (* Q: when are we using "when_born" and "conjecture_distance" of a clause? *)
    Clause.assign_when_born [c1] [c2] conclusion;
    let min_conj_dist = Clause.get_min_conjecture_distance [c1;c2] in
    Clause.assign_conjecture_distance (min_conj_dist+1) conclusion;  
    conclusion::rest  in 
  List.fold_left f [] c_list2     


let rec create_pred_map_cl_list clause_set =
	Set.fold add_clause_pred_map clause_set SymbMap.empty

let predElim clause_list =
	let clause_set = Clause.clause_list_to_set clause_list in 
	let entry = ref false in
	let predMap = create_pred_map_cl_list clause_set in
	(* let predList = Map.bindings predMap in *)
	let predList = SymbMap.fold (fun key _val rest -> (key::rest)) predMap [] in
	try
		while (not !entry) do
		    let f rest pred =
		    	let pred_info = SymbMap.find pred predMap in 
		    	if (pred_info.multi_occ_cnt == 0) then
			    let posClauses = pred_info.pos_cl_list in
			    let negClauses = pred_info.neg_cl_list in
			    (* build resolution resolvents *)
			    let r rest pos_cl =
			    	let curr_lit_list = Clause.find_all (fun l_pred -> ((Term.get_top_symb (Term.get_atom l_pred)) == pred)) pos_cl in
				let curr_lit = List.hd curr_lit_list in
				(* TODO: add a checker for curr_lit_list length *)
			    	let compl_curr_lit = Term.compl_lit curr_lit in 
			    	let r_to_add = resolution pos_cl curr_lit compl_curr_lit negClauses in
				r_to_add::rest in
			    List.fold_left r [] posClauses
			    (* use dedicated function for forward simplification *)
			    (* use dedicated function for backward simplification *)
			else [];
			(*should it be true??*)
			let conclusion = Clause.create [] in
			conclusion::rest in
		     List.fold_left f [] predList
		 done
	with 
		Not_found ->
			out_str ("Could not run\n");

